import dataclasses
import os
from dataclasses import dataclass
from functools import lru_cache
from typing import Any, Dict, List, Optional

import requests
from requests import Response
from requests.models import HTTPError

from airplane._version import __version__
from airplane.api.entities import PromptReviewers
from airplane.exceptions import InvalidEnvironmentException
from airplane.params import ParamTypes, SerializedParam, serialize_param
from airplane.types import JSONType


@dataclass(frozen=True)
class ClientOpts:
    """Client options for an APIClient."""

    api_host: str
    api_token: str
    env_id: str


class APIClient:
    """API Client to interact with the public Airplane API."""

    _api_host: str
    _headers: Dict[str, str]

    def __init__(self, opts: ClientOpts, version: str):
        self._api_host = opts.api_host
        self._headers = {
            "X-Airplane-Token": opts.api_token,
            "X-Airplane-Client-Kind": "sdk/python",
            "X-Airplane-Client-Version": version,
            "X-Airplane-Env-ID": opts.env_id,
        }

    def create_run(
        self,
        task_id: str,
        parameters: Optional[Dict[str, Any]] = None,
        env: Optional[Dict[str, Any]] = None,
        constraints: Optional[Dict[str, Any]] = None,
    ) -> str:
        """Creates an Airplane run with parameters, env and constraints from a task id.

        Args:
            task_id: The id of the task to run.
            parameters: Optional map of parameter slugs to values.
            env: Optional map of environment variables.
            constraints: Optional map of run constraints.

        Returns:
            The id of the run.

        Raises:
            HTTPError: If the run cannot be created or executed properly.
        """
        resp = requests.post(
            f"{self._api_host}/v0/runs/create",
            json={
                "taskID": task_id,
                "params": parameters,
                "env": env or {},
                "constraints": constraints or {},
            },
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()["runID"]

    def execute_task(
        self,
        slug: str,
        param_values: Optional[Dict[str, ParamTypes]] = None,
        resources: Optional[Dict[str, str]] = None,
    ) -> str:
        """Executes an Airplane task with parameters and resources from a task slug.

        Args:
            slug: The slug of the task to run.
            param_values: Optional map of parameter slugs to values.
            resources: Optional map of resource aliases to ids.

        Returns:
            The id of the run.

        Raises:
            HTTPError: If the run cannot be executed.
        """
        serialized_params = {}
        for key, val in (param_values or {}).items():
            serialized_params[key] = serialize_param(val)
        resp = requests.post(
            f"{self._api_host}/v0/tasks/execute",
            json={
                "slug": slug,
                "paramValues": serialized_params,
                "resources": resources or {},
            },
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()["runID"]

    def get_run(self, run_id: str) -> Dict[str, Any]:
        """Fetches an Airplane run.

        Args:
            run_id: The id of the run to fetch.

        Returns:
            The Airplane run's attributes.

        Raises:
            HTTPError: If the run cannot be fetched.
        """
        resp = requests.get(
            f"{self._api_host}/v0/runs/get",
            params={"id": run_id},
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()

    def get_run_output(self, run_id: str) -> Any:
        """Fetches an Airplane's run output.

        Args:
            run_id: The id of the run for which to fetch output.

        Returns:
            The Airplane run's outputs.

        Raises:
            HTTPError: If the run outputs cannot be fetched.
        """
        resp = requests.get(
            f"{self._api_host}/v0/runs/getOutputs",
            params={"id": run_id},
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()["output"]

    def create_text_display(self, content: str) -> str:
        """Creates a text display.

        Args:
            content: Content to display

        Returns:
            The Airplane display's id.

        Raises:
            HTTPError: If the display could not be created.
        """
        resp = requests.post(
            f"{self._api_host}/v0/displays/create",
            json={"display": {"content": content, "kind": "markdown"}},
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()["id"]

    def create_json_display(
        self,
        payload: JSONType,
    ) -> str:
        """Creates a json display.

        Args:
            payload: Payload to display

        Returns:
            The Airplane display's id.

        Raises:
            HTTPError: If the display could not be created.
        """
        resp = requests.post(
            f"{self._api_host}/v0/displays/create",
            json={"display": {"value": payload, "kind": "json"}},
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()["id"]

    def create_table_display(
        self,
        columns: Optional[List[Dict[str, Optional[str]]]],
        rows: List[Dict[str, Any]],
    ) -> str:
        """Creates a table display.

        Args:
            columns: Table columns containing keys slug and name
            rows: Table rows containing a map from column name to value.

        Returns:
            The Airplane display's id.

        Raises:
            HTTPError: If the display could not be created.
        """
        resp = requests.post(
            f"{self._api_host}/v0/displays/create",
            json={"display": {"columns": columns, "rows": rows, "kind": "table"}},
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()["id"]

    def create_prompt(
        self,
        parameters: List[SerializedParam],
        reviewers: Optional[PromptReviewers],
        confirm_text: Optional[str],
        cancel_text: Optional[str],
        description: Optional[str],
    ) -> str:
        """Creates an Airplane prompt.

        Args:
            parameters: List of parameters.
            reviewers: Reviewers that are allowed to approve the prompt.
            confirm_text: Text of the confirmation button on the prompt dialog.
            cancel_text: Text of the cancellation button on the prompt dialog.
            description: Prompt description to display. Supports markdown.

        Raises:
            HTTPError: If the prompt cannot be created properly.
        """
        resp = requests.post(
            f"{self._api_host}/v0/prompts/create",
            json={
                "schema": {
                    "parameters": [dataclasses.asdict(p) for p in parameters],
                },
                "reviewers": {
                    "users": reviewers.users,
                    "groups": reviewers.groups,
                    "allowSelfApproval": reviewers.allow_self_approvals,
                }
                if reviewers
                else None,
                "confirmText": confirm_text,
                "cancelText": cancel_text,
                "description": description,
            },
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()["id"]

    def get_prompt(self, prompt_id: str) -> Dict[str, Any]:
        """Fetches an Airplane prompt.

        Args:
            prompt_id: The id of the prompt to fetch.

        Returns:
            The Airplane prompt's attributes.

        Raises:
            HTTPError: If the prompt cannot be fetched.
        """
        resp = requests.get(
            f"{self._api_host}/v0/prompts/get",
            params={"id": prompt_id},
            headers=self._headers,
        )
        self.__maybe_error_on_response(resp)
        return resp.json()["prompt"]

    @classmethod
    def __maybe_error_on_response(cls, resp: Response) -> None:
        if resp.status_code >= 400:
            raise HTTPError(resp.json()["error"])


def client_opts_from_env() -> ClientOpts:
    """Creates ClientOpts from environment variables.

    Returns:
        Unvalidated ClientOpts from environment variables.

    Raises:
         InvalidEnvironmentException: If the environment is improperly configured.
    """

    opts = ClientOpts(
        api_host=os.getenv("AIRPLANE_API_HOST", ""),
        api_token=os.getenv("AIRPLANE_TOKEN", ""),
        env_id=os.getenv("AIRPLANE_ENV_ID", ""),
    )
    if any(not x for x in [opts.api_host, opts.api_token, opts.env_id]):
        raise InvalidEnvironmentException()
    return opts


def api_client_from_env() -> APIClient:
    """Creates an APIClient from environment variables.

    Returns:
        An APIClient to interact with the Airplane API.

    Raises:
        InvalidEnvironmentException: If the environment is improperly configured.
    """
    return api_client(client_opts_from_env())


@lru_cache(maxsize=None)
def api_client(opts: ClientOpts) -> APIClient:
    """Creates an APIClient

    Args:
        opts: The ClientOpts to use for the APIClient.

    Returns:
        An APIClient to interact with the Airplane API.
    """
    return APIClient(opts, __version__)



import json
import os
from typing import Any, Dict

from airplane.exceptions import (
    InvalidEnvironmentException,
    UnknownResourceAliasException,
)

__AIRPLANE_RESOURCES_ENV_VAR = "AIRPLANE_RESOURCES"
__AIRPLANE_RESOURCES_VERSION_ENV_VAR = "AIRPLANE_RESOURCES_VERSION"


def __convert_resource_alias_to_id(alias: str) -> str:
    """Converts a resource alias to a resource id.

    Args:
        alias: The resource alias to convert.

    Returns:
        The resource id that the provided alias corresponds to.

    Raises:
        InvalidEnvironmentException: If the environment does not contain resources information.
        UnknownResourceAliasException: If the resource alias cannot be found or converted.
        RunTerminationException: If the run fails or is cancelled.
    """

    resources_version = os.environ.get(__AIRPLANE_RESOURCES_VERSION_ENV_VAR, "")
    if resources_version != "2":
        raise InvalidEnvironmentException

    try:
        resources: Dict[str, Dict[str, Any]] = json.loads(
            os.environ.get(__AIRPLANE_RESOURCES_ENV_VAR, "{}")
        )
    except json.JSONDecodeError as decode_error:
        raise InvalidEnvironmentException from decode_error

    if alias not in resources:
        raise UnknownResourceAliasException(alias)

    if "id" not in resources[alias]:
        raise InvalidEnvironmentException

    return resources[alias]["id"]



from typing import Any, Dict, List, Optional

import backoff
import deprecation
import requests

from airplane._version import __version__
from airplane.api.client import api_client_from_env
from airplane.api.entities import PromptReviewers, Run, RunStatus
from airplane.exceptions import (
    PromptPendingException,
    RunPendingException,
    RunTerminationException,
)
from airplane.params import ParamTypes, SerializedParam


def execute(
    slug: str,
    param_values: Optional[Dict[str, ParamTypes]] = None,
    resources: Optional[Dict[str, Any]] = None,
) -> Run:
    """Standard executes an Airplane task, waits for execution, and returns run metadata.

    Args:
        slug: The slug of the task to run.
        param_values: Optional map of parameter slugs to values.
        resources: Optional map of resource aliases to ids.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the task cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    client = api_client_from_env()
    run_id = client.execute_task(slug, param_values, resources)
    run_info = __wait_for_run_completion(run_id)
    outputs = client.get_run_output(run_id)
    # pylint: disable=redefined-outer-name
    run = Run(
        id=run_info["id"],
        task_id=run_info.get("taskID", None),
        param_values=run_info["paramValues"],
        status=RunStatus(run_info["status"]),
        output=outputs,
    )

    if run.status in {RunStatus.FAILED, RunStatus.CANCELLED}:
        raise RunTerminationException(run)

    return run


@deprecation.deprecated(
    deprecated_in="0.3.2",
    current_version=__version__,
    details="Use execute(slug, param_values) instead.",
)
def run(
    task_id: str,
    parameters: Optional[Dict[str, Any]] = None,
    env: Optional[Dict[str, Any]] = None,
    constraints: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Creates an Airplane run, waits for execution, and returns its output and status.

    Args:
        task_id: The id of the task to run.
        parameters: Optional map of parameter slugs to values.
        env: Optional map of environment variables.
        constraints: Optional map of run constraints.

    Returns:
        The status and outputs of the run.

    Raises:
        HTTPError: If the run cannot be created or executed properly.
    """
    client = api_client_from_env()
    run_id = client.create_run(task_id, parameters, env, constraints)
    run_info = __wait_for_run_completion(run_id)
    outputs = client.get_run_output(run_id)
    return {"status": run_info["status"], "outputs": outputs}


@backoff.on_exception(
    lambda: backoff.expo(factor=0.1, max_value=5),
    (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
        RunPendingException,
    ),
)
def __wait_for_run_completion(run_id: str) -> Dict[str, Any]:
    client = api_client_from_env()
    run_info = client.get_run(run_id)
    if run_info["status"] in ("NotStarted", "Queued", "Active"):
        raise RunPendingException()
    return run_info


def prompt_background(
    serialized_params: List[SerializedParam],
    *,
    reviewers: Optional[PromptReviewers] = None,
    confirm_text: Optional[str] = None,
    cancel_text: Optional[str] = None,
    description: Optional[str] = None,
) -> str:
    """Creates a prompt in the background, returning the prompt ID."""

    client = api_client_from_env()
    return client.create_prompt(
        parameters=serialized_params,
        reviewers=reviewers,
        confirm_text=confirm_text,
        cancel_text=cancel_text,
        description=description,
    )


@backoff.on_exception(
    lambda: backoff.expo(factor=0.1, max_value=5),
    (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
        PromptPendingException,
    ),
)
def wait_for_prompt(prompt_id: str) -> Dict[str, Any]:
    """Waits until a prompt is submitted and returns the prompt values."""
    client = api_client_from_env()
    prompt_info = client.get_prompt(prompt_id)
    if not prompt_info["submittedAt"]:
        raise PromptPendingException()
    return prompt_info



from __future__ import annotations

from dataclasses import dataclass
from textwrap import dedent
from typing import Optional

from airplane.api.entities import Run


class RunPendingException(Exception):
    """Exception that indicates a run is still in pending state."""


class PromptPendingException(Exception):
    """Exception that indicates a prompt is still in pending state."""


class InvalidEnvironmentException(Exception):
    """Exception that indicates an improperly configured environment."""

    def __str__(self) -> str:
        return "This task must be run inside of the Airplane runtime."


@dataclass
class UnknownResourceAliasException(Exception):
    """Exception that indicates a resource alias is unattached."""

    alias: str

    def __str__(self) -> str:
        return f"The resource alias {self.alias} is unknown (have you attached the resource?)."


@dataclass
class RunTerminationException(Exception):
    """Exception that indicates a run failed or was cancelled."""

    run: Run

    def __str__(self) -> str:
        if isinstance(self.run.output, dict) and isinstance(
            self.run.output.get("error"), str
        ):
            return self.run.output["error"]
        return f"Run {str(self.run.status.value).lower()}"


@dataclass
class InvalidAnnotationException(Exception):
    """Exception that indicates an invalid annotation was provided in task definition."""

    param_name: str
    prefix: str
    func_name: Optional[str] = None

    def __str__(self) -> str:
        source = f" from function `{self.func_name}`" if self.func_name else ""
        return dedent(
            f"""
            {self.prefix} for parameter `{self.param_name}`{source}.

            Type must be one of (str, int, float, bool, datetime.date, datetime.datetime,
            airplane.LongText, airplane.File, airplane.ConfigVar, airplane.SQL,
            Optional[T], or Annotated[T, airplane.ParamConfig(...)]).
            """
        )


class UnsupportedDefaultTypeException(Exception):
    """Exception that indicates a default value isn't supported for a given type."""


class InvalidTaskConfigurationException(Exception):
    """Exception that indicates an inline task configuration is invalid."""



import json
import uuid
from typing import Any, Iterable, Union

import deprecation

from airplane._version import __version__

_CHUNK_SIZE = 8192


def set_output(value: Any, *path: Union[str, int]) -> None:
    """Sets the task output with optional subpath.

    Args:
        value: The value to output.
        path: Variadic parameter that denotes the subpath of the output.
    """
    val = __json_dumps(value)
    __chunk_print(f"airplane_output_set{__to_output_path(path)} {val}")


def append_output(value: Any, *path: Union[str, int]) -> None:
    """Appends to an array in the task output with optional subpath.

    Args:
        value: The value to output.
        path: Variadic parameter that denotes the subpath of the output.
    """
    val = __json_dumps(value)
    __chunk_print(f"airplane_output_append{__to_output_path(path)} {val}")


@deprecation.deprecated(
    deprecated_in="0.3.0",
    current_version=__version__,
    details="Use append_output(value) instead.",
)
def write_output(value: Any) -> None:
    """Writes the value to the task's output.

    Args:
        value: The value to output.
    """
    val = __json_dumps(value)
    __chunk_print(f"airplane_output {val}")


@deprecation.deprecated(
    deprecated_in="0.3.0",
    current_version=__version__,
    details="Use append_output(value, name) instead.",
)
def write_named_output(name: str, value: Any) -> None:
    """Writes the value to the task's output, tagged by the key.

    Args:
        name: The identifier to tag the output with.
        value: The value to output.
    """
    val = __json_dumps(value)
    __chunk_print(f'airplane_output:"{name}" {val}')


def __to_output_path(path: Iterable[Union[str, int]]) -> str:
    ret = "".join([f"[{json.dumps(item)}]" for item in path])
    return "" if ret == "" else f":{ret}"


def __chunk_print(output: str) -> None:
    if len(output) <= _CHUNK_SIZE:
        print(output)
        return

    chunk_key = str(uuid.uuid4())
    for start in range(0, len(output), _CHUNK_SIZE):
        print(f"airplane_chunk:{chunk_key} {output[start:start+_CHUNK_SIZE]}")
    print(f"airplane_chunk_end:{chunk_key}")


def __json_dumps(value: Any) -> str:
    # The backend can't handle NaNs or Infs, so we have to convert these to null
    # values. It's kind of messy to do this out-of-the-box in Python, but we can
    # get it working the following way:
    #
    # We try to dump using allow_nan=False. If that fails, then we catch the
    # resulting ValueError and then dump allowing NaNs but parse it while
    # converting NaNs to None via parse_constant, before re-dumping it.
    try:
        return json.dumps(value, separators=(",", ":"), allow_nan=False)
    except ValueError:
        json_str = json.dumps(value, separators=(",", ":"))
        json_with_nones = json.loads(json_str, parse_constant=lambda constant: None)
        return json.dumps(json_with_nones, separators=(",", ":"))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['airplane', 'airplane.api', 'airplane.builtins', 'airplane.runtime']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=1.10.0,<2.0.0',
 'deprecation>=2.1.0,<3.0.0',
 'docstring-parser>=0.14.1,<0.15.0',
 'inflection>=0.5.1,<0.6.0',
 'python-slugify>=6.1.2,<7.0.0',
 'requests>=2.25.1,<3.0.0',
 'typing-extensions>=4.3.0,<5.0.0']

setup_kwargs = {
    'name': 'airplanesdk',
    'version': '0.3.16',
    'description': 'A Python SDK for writing Airplane tasks',
    'long_description': '# Airplane Python SDK [![PyPI](https://img.shields.io/pypi/v/airplanesdk)](https://pypi.org/project/airplanesdk/) [![PyPI - License](https://img.shields.io/pypi/l/airplanesdk)](./LICENSE) [![Docs](https://img.shields.io/badge/Docs-airplane-blue)](https://docs.airplane.dev/creating-tasks/python)\n\nSDK for writing [Airplane](https://airplane.dev) tasks in Python.\n\nTo learn more, see [the Airplane SDK docs](https://docs.airplane.dev/tasks/python-sdk).\n',
    'author': 'Airplane',
    'author_email': 'support@airplane.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://airplane.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)
